#include <cctype>
#include <cassert>
#include <cstring>
#include <iostream>
#include <string>
#include <cmath>
#include <cfenv>
#include "ExprParser.h"

const double PI = 3.14159265358979323846;
const double E = 2.7182818284590452354;

using namespace std;

// Table of standard functions
typedef double (*FuncPtr)(double);
//... typedef double (*FuncPtr2)(double, double); // Not used now
typedef pair<FuncPtr, int> FuncDsc;
//... typedef pair<void*, int> FuncDsc;

map<string, FuncDsc> functions = {
    { "sin", { &sin, 1 } },
    { "cos", { &cos, 1 } },
    { "tan", { &tan, 1 } },
    { "tg", { &tan, 1 } },
    { "exp", { &exp, 1 } },
    { "log", { &log, 1 } },
    { "ln", { &log, 1 } },
    { "atan", { &atan, 1 } },
    { "arctg", { &atan, 1 } },
    { "asin", {&asin, 1 } },
    { "arcsin", {&asin, 1 } },
    { "acos", {&acos, 1 } },
    { "arccos", {&acos, 1 } },
    { "sqrt", { &sqrt, 1 } },
    { "fabs", { &fabs, 1 } },
    { "abs", { &fabs, 1 } }
    // { "atan2", { reinterpret_cast<FuncPtr>(&atan2), 2 } }
};

std::ostream& operator<<(std::ostream& s, const Token& token) {
    if (token.lex == Token::NUMBER) {
        s << "NUMBER: " << token.number;
    } else if (token.lex == Token::PLUS) {
        s << "PLUS";
    } else if (token.lex == Token::MINUS) {
        s << "MINUS";
    } else if (token.lex == Token::MUL) {
        s << "MUL";
    } else if (token.lex == Token::DIV) {
        s << "DIV";
    } else if (token.lex == Token::POW) {
        s << "POW";
    } else if (token.lex == Token::LPAR) {
        s << "LPAR";
    } else if (token.lex == Token::RPAR) {
        s << "RPAR";
    } else if (token.lex == Token::X) {
        s << "X";
    } else if (token.lex == Token::Y) {
        s << "Y";
    } else if (token.lex == Token::NAME) {
        s << "NAME: " << token.name;
    } else if (token.lex == Token::COMMA) {
        s << "COMMA";
    } else if (token.lex == Token::UMINUS) {
        s << "UMINUS";
    } else if (token.lex == Token::ENDLINE) {
        s << "ENDLINE";
    } else if (token.lex == Token::ILLEGAL) {
        s << "ILLEGAL";
    } else {
        assert(false);
    }
    return s;
}

void Scanner::init(const char* input) {
    inputLine = input;
    curpos = 0;
    len = int(strlen(input));
    lexemeReady = false;
    lexeme = Token();
}

Token Scanner::getToken() {
    if (!lexemeReady) {
        readToken();
    }
    if (!lexemeReady)
       return Token(Token::ENDLINE);
    lexemeReady = false;
    return lexeme;
}

const Token& Scanner::peekToken() {
    if (!lexemeReady) {
        readToken();
    }
    return lexeme;
}

void Scanner::readToken() {
    assert(!lexemeReady);

    // Skip white space
    while (curpos < len && isspace(inputLine[curpos]))
        ++curpos;
    if (curpos >= len) {
        lexeme = Token(Token::ENDLINE);
        return;
    }
    
    if (inputLine[curpos] == '+') {
        lexeme = Token(Token::PLUS);
        ++curpos;
    } else if (inputLine[curpos] == '-') {
        lexeme = Token(Token::MINUS);
        ++curpos;
    } else if (inputLine[curpos] == '*') {
        lexeme = Token(Token::MUL);
        ++curpos;
    } else if (inputLine[curpos] == '/') {
        lexeme = Token(Token::DIV);
        ++curpos;
    } else if (inputLine[curpos] == '^') {
        lexeme = Token(Token::POW);
        ++curpos;
    } else if (inputLine[curpos] == '(') {
        lexeme = Token(Token::LPAR);
        ++curpos;
    } else if (inputLine[curpos] == ')') {
        lexeme = Token(Token::RPAR);
        ++curpos;
    } else if (inputLine[curpos] == ',') {
        lexeme = Token(Token::COMMA);
        ++curpos;
    } else if (isdigit(inputLine[curpos])) {
        // 0.5 in correct, but .5 is not correct
        string number;
        number += inputLine[curpos];
        ++curpos;
        int numPoints = 0;
        while (
            curpos < len &&
            numPoints <= 1 &&
            (isdigit(inputLine[curpos]) ||
            inputLine[curpos] == '.')
        ) {
            if (inputLine[curpos] == '.') {
                if (numPoints == 1) {
                    break;
                } else {                 
                    ++numPoints;
                }
            }
            number += inputLine[curpos];
            ++curpos;
        }     
        lexeme = Token(Token::NUMBER, std::stod(number));
    } else if (isalpha(inputLine[curpos])) {
        string name;
        name += inputLine[curpos];
        ++curpos;
        while (
            curpos < len &&
            (isalnum(inputLine[curpos]) ||
            inputLine[curpos] == '_')
        ) {
            name += inputLine[curpos];
            ++curpos;
        }     
        if (name == "x" || name == "X") {
            lexeme = Token(Token::X, 0., "x");
        } else if (name == "y" || name == "Y") {
            lexeme = Token(Token::Y, 0., "y");
        } else if (name == "pi" || name == "PI") {
            lexeme = Token(Token::NUMBER, PI, "pi");
        } else if (name == "e" || name == "E") {
            lexeme = Token(Token::NUMBER, E, "e");
        } else {
            lexeme = Token(Token::NAME, 0., name);
        }
    } else {
        ++curpos;
        lexeme = Token(Token::ILLEGAL);
    } 
    lexemeReady = true; 
}

double Rpn::evaluate(double x, double y) const {
    fenv_t fenv;
    int float_point_exceptions = feholdexcept(&fenv);
    // std::cout << "feholdexcept(&fenv) returns "
    //           << float_point_exceptions << std::endl;
    int res = std::feclearexcept(FE_ALL_EXCEPT);
    if (res != 0) {
        std::cout << "feclearexcept(FE_ALL_EXCEPT) returns "
                  << res << std::endl;
    }

    stack.clear();
    for (size_t i = 0; i < size(); ++i) {
        double u, v;
        const Token& token = (*this)[i];
        if (token.lex == Token::NUMBER) {
            stack.push_back(token.number);
        } else if (token.lex == Token::PLUS) {
            v = stack.back(); stack.pop_back();
            u = stack.back(); stack.pop_back();
            stack.push_back(u + v);
        } else if (token.lex == Token::MINUS) {
            v = stack.back(); stack.pop_back();
            u = stack.back(); stack.pop_back();
            stack.push_back(u - v);
        } else if (token.lex == Token::MUL) {
            v = stack.back(); stack.pop_back();
            u = stack.back(); stack.pop_back();
            stack.push_back(u * v);
        } else if (token.lex == Token::DIV) {
            v = stack.back(); stack.pop_back();
            u = stack.back(); stack.pop_back();
            stack.push_back(u / v);
        } else if (token.lex == Token::UMINUS) {
            stack.back() = (-stack.back());
        } else if (token.lex == Token::POW) {
            v = stack.back(); stack.pop_back();
            u = stack.back(); stack.pop_back();
            stack.push_back(pow(u, v));
        } else if (token.lex == Token::X) {
            stack.push_back(x);
        } else if (token.lex == Token::Y) {
            stack.push_back(y);
        } else if (token.lex == Token::NAME) {
            if (functions.count(token.name) == 1) {
                auto fDsc = functions[token.name];
                if (fDsc.second == 1) {
                    u = stack.back(); stack.pop_back();
                    stack.push_back(
                        (*(static_cast<FuncPtr>(fDsc.first))) (u)
                    );
                } 
                /*... Not used now
                else if (fDsc.second == 2) {
                    v = stack.back(); stack.pop_back();
                    u = stack.back(); stack.pop_back();
                    stack.push_back(
                        (*(reinterpret_cast<FuncPtr2>(fDsc.first))) (u, v)
                        // (*(FuncPtr2)(fDsc.first))(u, v)
                    );
                } 
                ...*/
                else {
                    assert(false);
                }
            }
        }
    }
    if (float_point_exceptions == 0) {
        int excepts = (
            FE_DIVBYZERO | FE_INVALID
            // FE_UNDERFLOW | FE_OVERFLOW
        );
        int res = std::fetestexcept(excepts);
        // std::cout << "fetestexcept(excepts) = "
        //           << res << std::endl;
        if (res != 0) {
            // std::cout << "Floating point exception"
            //           << std::endl;
            std::feclearexcept(FE_ALL_EXCEPT);
            throw ParserException("Floating point exception");
        }
    }
    return stack.back();    
}

void Parser::init(const char* input) {
    scanner.init(input);
    rpn.clear();
}
    
bool Parser::parse(std::string& message) {
    rpn.clear();
    try {
        processS();
        return true;
    } catch (const ParserException& e) {
        message = e.what();
        return false;
    }        
}

void Parser::processS() {
    processF();
    skipToken(Token::ENDLINE);
}

void Parser::processF() {
    processT();
    processE();
}

void Parser::processE() {
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::PLUS) {
        skipToken(Token::PLUS);
        processT(); 
        rpn.push_back(Token::PLUS);
        processE();
    } else if (nextToken.lex == Token::MINUS) {
        skipToken(Token::MINUS);
        processT(); 
        rpn.push_back(Token::MINUS);
        processE();
    } else {
        // Empty sequence, nothing to do
    }
}

void Parser::processT() {
    processP();
    processG();
}

void Parser::processG() {
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::MUL) {
        skipToken(Token::MUL);
        processP(); 
        rpn.push_back(Token::MUL);
        processG();
    } else if (nextToken.lex == Token::DIV) {
        skipToken(Token::DIV);
        processP(); 
        rpn.push_back(Token::DIV);
        processG();
    } else {
        // Empty sequence, nothing to do
    }
}

void Parser::processP() {
    bool changeSign = false;
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::MINUS) {
        skipToken(Token::MINUS);
        changeSign = true;
    }
    processM();
    if (changeSign) {
        rpn.push_back(Token::UMINUS);
    }
    processK();
}

void Parser::processK() {
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::POW) {
        skipToken(Token::POW);
        processM(); 
        processK();
        rpn.push_back(Token::POW);
    } else {
        // Empty sequence, nothing to do
    }
}

void Parser::processM() {
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::NUMBER) {
        rpn.push_back(nextToken);
        skipToken(Token::NUMBER);
    } else if (nextToken.lex == Token::X) {
        rpn.push_back(nextToken);
        skipToken(Token::X);
    } else if (nextToken.lex == Token::Y) {
        rpn.push_back(nextToken);
        skipToken(Token::Y);
    } else if (nextToken.lex == Token::LPAR) {
        skipToken(Token::LPAR);
        processF();
        skipToken(Token::RPAR);
    } else if (nextToken.lex == Token::NAME) {
        skipToken(Token::NAME);
        skipToken(Token::LPAR);
        processArgs();
        skipToken(Token::RPAR);
        rpn.push_back(nextToken); // Function name
    } else {
        throw ParserException("Syntax error");
    }
}

void Parser::processArgs() {
    processF();
    processH();
}

void Parser::processH() {
    Token nextToken = scanner.peekToken();
    if (nextToken.lex == Token::COMMA) {
        skipToken(Token::COMMA);
        processF();
    } else {
        // Empty, nothing to do
    }
}

void Parser::skipToken(int token) {
    Token t = scanner.getToken();
    if (t.lex != token) 
        throw ParserException("Syntax error");
}

double Parser::evaluate(double x, double y) const {
    return rpn.evaluate(x, y);
}
